#!/usr/bin/env python
# Copyright (C) 2016 Luke Kenneth Casson Leighton <lkcl@lkcl.net>
# License: GPLv3+


import time
import math
import os
import itertools

import openscad

from pyopenscad import *
from polysurface import spline_double_divide, Offset, SurfaceFromCurves
from polysurface import vtrans, vmult, vsub
from copy import deepcopy
from pyscadutil import B0, B45, pin
from pyscadutil import curvebox, shearedcube, hollowbox, touch_arc
from pyscadutil import L, R, T, D, F, B

from pyopenscadobj import * # XXX don't do this!


def add_boltset(obj, spec):
        for holesize, kls, pos, rot, length in spec:
            if isinstance(holesize, int) or isinstance(holesize, float):
                c = cylinder(r=holesize/2, h=length, center=True, segments=20)
            else:
                c = holesize
            if rot and not (isinstance(rot[0], list) \
                            or isinstance(rot[0], tuple)):
                rot = [rot]
            for r in rot:
                c = rotate(r).add(c)
            c = translate(pos).add(c)
            obj.o -= c
            if kls:
                b = kls(length)
                b.rotate(rot)
                b.set_parent(obj, pos)


class Bolt(Object):
    def __init__(self, diameter, length, extra_diam=0.0,
                        add_nut=False, allenkey=False,
                        pozidrive=False, countersunk=False, selftap=False,
                        lowprofile=False, hexshort=False):

        self.length = length
        if allenkey:
            if lowprofile:
                hdlen = diameter*0.3
            else:
                hdlen = diameter*0.8
            hdiam = diameter*1.6
            ndiam = diameter*2.2
            rounding=20
            self.typ = "Allen"
        elif selftap:
            hdlen = diameter*0.7
            hdiam = diameter*1.8
            ndiam = diameter*2.2
            rounding=20
            self.typ = "SelfTap"
        elif pozidrive:
            hdlen = diameter*0.7
            hdiam = diameter*1.8
            ndiam = diameter*2.2
            rounding=20
            self.typ = "Pozi"
        elif hexshort:
            hdlen = diameter*0.5
            hdiam = diameter*1.50
            ndiam = diameter*2.2
            rounding=6
            self.typ = "Hex"
        else:
            hdlen = diameter*0.6
            hdiam = diameter*2.2
            ndiam = diameter*2.2
            rounding=6
            self.typ = "Hex"
        u = union()
        if countersunk:
            csdiam = diameter/2.0
        else:
            csdiam = hdiam * 0.6

        # after above, adjust the diameter (if any)
        diameter += extra_diam
        csdiam += extra_diam

        c = cylinder(r=hdiam*0.6, r1=csdiam,
                     h=hdlen+0.01, center=True, segments=rounding)
        u += translate([0, 0, (length+hdlen)/2]).add(c)
        u += cylinder(r=diameter/2.0, h=length, center=True, segments=20)
        if add_nut:
            c = cylinder(r=ndiam*0.4, h=hdlen, center=True, segments=6)
            u += translate([0, 0, (-length+hdlen+5)/2]).add(c)
            c = cylinder(r=ndiam*0.4, r1=ndiam*0.25,
                            h=hdlen/2, center=True, segments=6)
            u += translate([0, 0, (-length-hdlen*0.5+5)/2]).add(c)
        Object.__init__(self, u, BOM_PART|BOM_NOSCAD)

    def name(self):
        return "%s %s (%dmm)" % (self.__class__.__name__, self.typ, self.length)


class M25HexBolt(Bolt):
    def __init__(self, length, add_nut=True):
        Bolt.__init__(self, 2.5, length, add_nut=add_nut)

class M3HexBolt(Bolt):
    def __init__(self, length, add_nut=True):
        Bolt.__init__(self, 3.0, length, add_nut=add_nut)


class M4HexBolt(Bolt):
    def __init__(self, length, add_nut=True, countersunk=False):
        Bolt.__init__(self, 4.0, length, add_nut=add_nut, 
                            countersunk=countersunk)


class M4PoziHead(Bolt):
    def __init__(self, length, add_nut=True, countersunk=False):
        Bolt.__init__(self, 4.0, length, add_nut=add_nut, pozidrive=True,
                            countersunk=countersunk)


class M2PoziHead(Bolt):
    def __init__(self, length, add_nut=True):
        Bolt.__init__(self, 2.0, length, add_nut=add_nut, pozidrive=True)


# M3 bolts

class M3PoziHead(Bolt):
    def __init__(self, length, add_nut=True, countersunk=False,
                               extra_diam=0.0):
        Bolt.__init__(self, 3.0, length, add_nut=add_nut, pozidrive=True,
                            countersunk=countersunk, extra_diam=extra_diam)


class M3SelfTapping(Bolt):
    def __init__(self, length):
        Bolt.__init__(self, 3.0, length, selftap=True)

class M3AllenBolt(Bolt):
    def __init__(self, length, add_nut=True):
        Bolt.__init__(self, 3.0, length, add_nut=add_nut, allenkey=True)

# M5 bolts

class M5AllenKey(Bolt):
    def __init__(self, length, add_nut=True):
        Bolt.__init__(self, 5.0, length, add_nut=add_nut, allenkey=True)

class M5HexBolt(Bolt):
    def __init__(self, length, add_nut=True):
        Bolt.__init__(self, 5.0, length, add_nut=add_nut)

class M5PoziHead(Bolt):
    def __init__(self, length, add_nut=True, countersunk=False,
                               extra_diam=0.0):
        Bolt.__init__(self, 5.0, length, add_nut=add_nut, pozidrive=True,
                            countersunk=countersunk, extra_diam=extra_diam)

class M5SmallHeadHex(Bolt):
    def __init__(self, length, add_nut=True):
        Bolt.__init__(self, 5.0, length, add_nut=add_nut, hexshort=True)


class M6SmallHeadHex(Bolt):
    def __init__(self, length, add_nut=True):
        Bolt.__init__(self, 6.0, length, add_nut=add_nut, hexshort=True)

class M6SmallHeadHex(Bolt):
    def __init__(self, length, add_nut=True):
        Bolt.__init__(self, 6.0, length, add_nut=add_nut, hexshort=True)

# M8 bolts

class M8PoziHead(Bolt):
    def __init__(self, length, add_nut=True):
        Bolt.__init__(self, 8.0, length, add_nut=add_nut, pozidrive=True)

class M8HexBolt(Bolt):
    def __init__(self, length, add_nut=True):
        Bolt.__init__(self, 8.0, length, add_nut=add_nut)

class M8CapScrew(Bolt):
    def __init__(self, length, add_nut=False):
        Bolt.__init__(self, 8.0, length, add_nut=add_nut, allenkey=True,
                            lowprofile=True)
        self.url = "https://www.nbk1560.com/en-US/products/specialscrew/nedzicom/lowsmallheadscrew/SSHS/"

class M8AllenKey(Bolt):
    def __init__(self, length, add_nut=True):
        Bolt.__init__(self, 8.0, length, add_nut=add_nut, allenkey=True)

class M8AllenKeyCounter(Bolt):
    def __init__(self, length, add_nut=True):
        Bolt.__init__(self, 8.0, length, add_nut=add_nut, allenkey=True,
                            countersunk=True)

class M8SmallHeadHex(Bolt):
    def __init__(self, length, add_nut=True):
        Bolt.__init__(self, 8.0, length, add_nut=add_nut, hexshort=True)

# M10 bolts

class M10HexBolt(Bolt):
    def __init__(self, length, add_nut=True):
        Bolt.__init__(self, 10.0, length, add_nut=add_nut)


class Nut(Object):
    def __init__(self, diam, h, idiam):
        self.diameter = diam
        self.height = h
        u = union()
        u += cylinder(diam/2.0, h=h, center=True, segments=6)
        u -= cylinder(idiam/2.0, h=h+1, center=True, segments=6)
        Object.__init__(self, u, bomtype=BOM_3D)


class NutM8x1(Nut):
    def __init__(self):
        Nut.__init__(self, 14, 6.5, 8.8)
        self.description = "DIN 935 M8x1 BZP 8.8 Hexagon nut"
        self.url = "http://alfa-tech3d.com/en/nuts/"\
                    "hexagon-nut-din-934-88-bzp-m8x1f.html"


class TR8x2(Object):
    def __init__(self):
        u = union()
        u += cylinder(9.9/2.0, h=15, segments=20)
        c = cylinder(22/2.0, h=3.5, segments=30)
        u += translate([0, 0, 1.5]).add(c)
        c = cylinder(6.45/2.0, h=16, segments=20)
        u -= translate([0, 0, -1]).add(c)

        # screw-holes
        c = cylinder(3.55/2.0, h=30, center=True, segments=20)
        u -= translate([8, 0, 0]).add(c)
        u -= translate([-8, 0, 0]).add(c)
        u -= translate([0, -8, 0]).add(c)
        u -= translate([0, 8, 0]).add(c)

        Object.__init__(self, u, bomtype=BOM_3D,
            description="TR8x2 Trapezoidal Nut, 8mm x 2mm pitch",
            url="http://alfa-tech3d.com/en/nuts/trapezoidal-nut-tr8-x-2.html")

        self.set_colour([0.9, 0.7, 0.1, 0.4])
